// ****************************************************************************
// Copyright (C) 1998 by Dieter Windau
// All rights reserved
//
// BmpMenu.cpp: implementation file
// Version:     1.2
// Date:        07.08.1998
// Author:      Dieter Windau
//
// TBmpMenu is a freeware OWL class that shows bitmaps
// on the left side of menus like MS Office 97
//
// Portions of code are based on MFC class CMenuSpawn written by Iuri
// Apollonio and BCMenu written by Brent Corkum. Very special thanks.
//
// You are free to use/modify this code but leave this header intact.
// May not be sold for profit.
//
// Tested with Borland C++ 5.02, OWL 5.02 under Windows NT 4.0 SP3 but I think
// the class would work with Windows 95 too.
// This file is provided "as is" with no expressed or implied warranty.
// Use at your own risk.
//
// Please send me bug reports, bug fixes, enhancements, requests, etc., and
// I'll try to keep this in next versions:
//   EMail: dieter.windau@usa.net
//   Web:   http://www.members.aol.com/SoftEngage
// ****************************************************************************

#include <owl\owlpch.h>
#pragma hdrstop

#include "BmpMenu.h"

// TBmpMenu use the Win32 API and can't work with Windows 3.11
#if defined(BI_PLAT_WIN32)

const DefaultBmpMenuWidth  = 19;  // if you use default borland size,
const DefaultBmpMenuHeight = 18;  // change the values to 20 x 20

DIAG_DEFINE_GROUP(BmpMenu, 1, 0);

// ******************************* helper functions ***************************

void Draw3dRect(HDC hdc, const TRect& rect,
	COLORREF clrTopLeft, COLORREF clrBottomRight)
// MFC style function, that draw the frame of a 3d rect with given colors
{
	if (hdc == 0)
  	return;

	Draw3dRect(hdc, rect.left, rect.top, rect.Width(),
		rect.Height(), clrTopLeft, clrBottomRight);
}

void Draw3dRect(HDC hdc, int x, int y, int cx, int cy,
	COLORREF clrTopLeft, COLORREF clrBottomRight)
{
	if (hdc == 0)
  	return;

	FillSolidRect(hdc, x, y, cx - 1, 1, clrTopLeft);
	FillSolidRect(hdc, x, y, 1, cy - 1, clrTopLeft);
	FillSolidRect(hdc, x + cx, y, -1, cy, clrBottomRight);
	FillSolidRect(hdc, x, y + cy, cx, -1, clrBottomRight);
}

void FillSolidRect(HDC hdc, int x, int y, int cx, int cy, COLORREF clr)
// MFC style function, that draws a filled rect. Is this faster???
{
	if (hdc == 0)
  	return;

	::SetBkColor(hdc, clr);
	TRect rect(x, y, x + cx, y + cy);
	::ExtTextOut(hdc, 0, 0, ETO_OPAQUE, &rect, 0, 0, 0);
}

// ******************** TBmpMenu *********************************************

TBmpMenu::TBmpMenu(TModule* module):
  MenuItems(5,0,5),
  ImageItems(5,0,5),
  ExcludedBitmaps(5,0,5)
{
  Module = module ? module : ::Module;
  bShowAccInTooltips = bDrawRecessed = true;
  bIsPopup = bBackBitmap = false;

  // The size of the bitmaps on the left side are based on the
  // default bitmap size. Cut the right and bottom side, because
  // it's always gray. Change the constants at the top of
  // source, if you use default Borland size (20 x 20).
  //
	ImageSize = TSize(DefaultBmpMenuWidth, DefaultBmpMenuHeight);
  ImageSize.cx -= 1;
	ImageSize.cy -= 1;

  ImageList = new TImageList(ImageSize, ILC_COLOR4|ILC_MASK, 0, 1);
	CHECK(ImageList);
  Old3dFaceColor = TColor::Sys3dFace;
  ImageList->SetBkColor(Old3dFaceColor);

  BackBitmap = 0;
  TRACEX(BmpMenu, 0, "TBmpMenu constructed @" << (void*)this);
}

TBmpMenu::~TBmpMenu()
{
// is not necessary because the bug will be fixed in future > #2 patch
/*
#if (OWLInternalVersion >= 0x06000000L)
  uint i;
  for (i=MenuItems.GetItemsInContainer(); i>0; i--)
  {
    delete MenuItems[i-1];
    MenuItems[i-1] = 0;
  }
  for (i=ImageItems.GetItemsInContainer(); i>0; i--)
  {
    delete ImageItems[i-1];
    ImageItems[i-1] = 0;
  }
#endif
*/
  MenuItems.Flush();
  ImageItems.Flush();
  ImageList->RemoveAll();
  delete ImageList;
  delete BackBitmap;
  TRACEX(BmpMenu, 0, "TBmpMenu destructed @" << (void*)this);
}

void TBmpMenu::RemapMenu(HMENU hMenu, bool isPopup)
// Remap a simple menu created with Windows API functions or with OWL TMenu
// class to a menu with bitmaps on the left side. Should be called from e.g:
// SetupWindow(), EvInitMenu(), EvInitMenuPopup() functions and before the
// TrackPopupMenu() function.
// If the hMenu is a popupMenu set isPopup = true
{
	PRECONDITION(hMenu);
  TRACEX(BmpMenu, 0, "RemapMenu @" << (void*)hMenu << " isPopup=" << isPopup);
  SyncronizeAllMenuItems();
  if (::IsMenu(hMenu))
  {
    bIsPopup = isPopup;
	  RemapMenuRecursiv(hMenu);
  }
}

bool TBmpMenu::DrawItem(DRAWITEMSTRUCT far& DIS)
// To be called from the TWindow EvDrawItem(..)
// Call the base class if it return false
{
	bool res = false;
	if (DIS.CtlType == ODT_MENU)
	{
	  if (Old3dFaceColor != TColor::Sys3dFace)
    {
      // if 3dFace color change, change the background of all bitmaps in list
      //
      Old3dFaceColor = TColor::Sys3dFace;
	  	ImageList->SetBkColor(Old3dFaceColor);
    }

		UINT state = DIS.itemState;
		bool bEnab =  !(state & ODS_DISABLED);
		bool bSelect = (state & ODS_SELECTED) ? true : false;
		bool bChecked = (state & ODS_CHECKED) ? true : false;
    bool bItemInfoValid;
    bool bDefault;
    bool bRadioCheck;

    bItemInfoValid = bDefault = bRadioCheck = false;
		TBmpMenuItem* pItem = (TBmpMenuItem*)DIS.itemData;
		if (pItem)
		{
      // if it is not a submenu item, get the menu item info
      //
      TMenuItemInfo mii(MIIM_STATE | MIIM_TYPE);
      if (pItem->Cmd >= 0)
      {
        if ((bItemInfoValid = ::GetMenuItemInfo((HMENU)DIS.hwndItem, DIS.itemID,
    	   	MF_BYCOMMAND, &mii)) == true)
        {
         	if (mii.fState & MFS_DEFAULT)
	          bDefault = true;
          if (mii.fType & MFT_RADIOCHECK)
            bRadioCheck = true;
        }
      }

			TDC dc(DIS.hDC);

      // get the actual windows wide menu font
      //
      TFont* pft;
		  NONCLIENTMETRICS nm;
    	nm.cbSize = sizeof (NONCLIENTMETRICS);
    	if (SystemParametersInfo(SPI_GETNONCLIENTMETRICS, nm.cbSize, &nm, 0))
      {
        if (bItemInfoValid && bDefault)
          nm.lfMenuFont.lfWeight = FW_BOLD;
        pft = new TFont(&nm.lfMenuFont);
      }
      else
        pft = new TFont((HFONT)GetStockObject(DEFAULT_GUI_FONT));
      CHECK(pft);

			dc.SelectObject(*pft);
			TRect rc(DIS.rcItem);
			TRect rcImage(rc);
      TRect rcText(rc);
			rcImage.right = rcImage.left + rc.Height();
			rcImage.bottom = rc.bottom;

			if (bBackBitmap && BackBitmap && (HBITMAP)BackBitmap &&
          BackBitmap->Width() > 0 && BackBitmap->Height() > 0)
			{
        // draw background bitmap
        //
				TMemoryDC tempDC(dc);
				tempDC.FillRect(rc, TBrush(TColor::SysMenu));
				tempDC.SelectObject(*BackBitmap);
				dc.FillRect(rc, TBrush(TColor::SysMenu));

        // draw bitmap tiled
        //
        int xBegin = (int)(rc.left/BackBitmap->Width());
        int xEnd   = (int)(rc.right/BackBitmap->Width())+1;
        int yBegin = (int)(rc.top/BackBitmap->Height());
        int yEnd   = (int)(rc.bottom/BackBitmap->Height())+1;
        int i,j;
        for (i = xBegin; i < xEnd; i++)
          for (j = yBegin; j < yEnd; j++)
          {
	          tempDC.SetWindowOrg(TPoint(
  	          i*BackBitmap->Width(), j*BackBitmap->Height()));
	          dc.BitBlt(rc, tempDC, rc.TopLeft(), SRCCOPY);
          }
			}

			if (pItem->Cmd == -3) // is a separator
			{
				TPen pnDk(TColor::Sys3dShadow);
				TPen pnLt(TColor::Sys3dHilight);
				dc.SelectObject(pnDk);
				dc.MoveTo(rc.left + 2, rc.top + 2);
				dc.LineTo(rc.right - 2, rc.top + 2);
				dc.RestorePen();
				dc.SelectObject(pnLt);
				dc.MoveTo(rc.left + 2, rc.top + 3);
				dc.LineTo(rc.right - 2, rc.top + 3);
				dc.RestorePen();
			}
      else if (pItem->Cmd == -4) // ia a title item
			{
				string cs(pItem->Text);
				TRect rcBdr(rcText);

				if (bDrawRecessed)
        {
          // draw recessed title item
          //
          if (bSelect && bEnab)
          {
            rcText.top ++;
            rcText.left += 2;
          }

          if (!bBackBitmap)
            dc.FillRect(rcText, TBrush(TColor::SysMenu));
          dc.DrawText(cs.c_str(), -1, rcText, DT_VCENTER|DT_CENTER|DT_SINGLELINE);

          if (bSelect && bEnab)
            Draw3dRect(DIS.hDC, rcBdr, TColor::Sys3dShadow, TColor::Sys3dHilight);
        }
        else
        {
          // draw normal title item
         	//
          if (bSelect && bEnab)
          {
	          dc.FillRect(rcText, TBrush(TColor::SysHighlight));
            dc.SetTextColor(TColor::SysHighlightText);
            dc.SetBkColor(TColor::SysHighlight);
          }
          else
          {
	          if (!bBackBitmap)
	            dc.FillRect(rcText, TBrush(TColor::SysMenu));
          }
          dc.DrawText(cs.c_str(), -1, rcText, DT_VCENTER|DT_CENTER|DT_SINGLELINE);
        }
			}
      else
      {
        // draw menu item
        //
        rcText.left += rcImage.right + 1;
        int obk = dc.SetBkMode(TRANSPARENT);
        COLORREF ocr;

        if (bSelect)
        {
          if (pItem->ImageIdx >= 0 || bChecked)
          {
            dc.FillRect(rcText, TBrush(TColor::SysHighlight));
            if (bChecked)
              dc.FillRect(rcImage, TBrush(TColor::SysMenu));
          }
          else
            dc.FillRect(rc, TBrush(TColor::SysHighlight));
          ocr = dc.SetTextColor(TColor::SysHighlightText);
        }
        else
        {
          if (pItem->ImageIdx >= 0 || bChecked)
          {
            if (!bBackBitmap)
              dc.FillRect(rcText, TBrush(TColor::SysMenu));
            if (bChecked)
            {
              // draw 50% mask
              //
              rcImage.Inflate(-1, -1);
              const uint8 mask[8] =
                { 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55 };
              THatch8x8Brush br(mask, TColor::Sys3dHilight, TColor::Sys3dFace);
              dc.FillRect(rcImage, br);
              rcImage.Inflate(1, 1);
            }
          }
          else if (!bBackBitmap)
            dc.FillRect(rc, TBrush(TColor::SysMenu));
          ocr = dc.SetTextColor(TColor::SysMenuText);
        }

        int ay = (rcImage.Height() - ImageSize.cy) / 2;
        int ax = (rcImage.Width()  - ImageSize.cx) / 2;

        if (pItem->ImageIdx >= 0)
        {
          if (bSelect && bEnab)
          {
            if (bChecked)
              Draw3dRect(DIS.hDC, rcImage, TColor::Sys3dShadow, TColor::Sys3dHilight);
            else
              Draw3dRect(DIS.hDC, rcImage, TColor::Sys3dHilight, TColor::Sys3dShadow);
          }
          else
          {
            if (bChecked)
              Draw3dRect(DIS.hDC, rcImage, TColor::Sys3dShadow, TColor::Sys3dHilight);
            else
              Draw3dRect(DIS.hDC, rcImage, TColor::SysMenu, TColor::SysMenu);
          }

          ImageList->Draw(pItem->ImageIdx, dc, rcImage.left + ax,
            rcImage.top + ay, (bChecked) ? ILD_TRANSPARENT : ILD_NORMAL);

          if (!bEnab)
          {
            TRect strictImgRc(rcImage.left + ax, rcImage.top + ay,
              rcImage.left + ax + ImageSize.cx, rcImage.top + ay + ImageSize.cy);
            DrawDisabledButton(dc, strictImgRc);
          }
        }
        else
        {
          if (bChecked)
          {
            rcImage.Inflate(-1,-1);
            Draw3dRect(DIS.hDC, rcImage, TColor::Sys3dShadow, TColor::Sys3dHilight);

            if (bItemInfoValid && bRadioCheck)
              DrawRadioDot(dc, rcImage.left + ax + 5,
                rcImage.top + ay + 4, TColor::SysMenuText);
            else
              DrawCheckmark(dc, rcImage.left + ax + 5,
                rcImage.top + ay + 4, TColor::SysMenuText);
          }
        }

        string cs1;
        string cs(pItem->Text);
        TSize sz = dc.GetTextExtent(cs.c_str(), strlen(cs.c_str()));
        int ay1 = (rcText.Height() - sz.cy) / 2;
        rcText.top += ay1;
        rcText.left += 2;
        rcText.right -= 15;

        int tf = cs.find("\t");

        // \a in resource is a tab too
        //
        // if (tf == NPOS)
        //   tf = cs.find("\a");

	      // Accelerator can be \b instead of \t (NOTE: \a in .rc make a \b in res)
    	  //
        if (tf == NPOS)
          tf = cs.find("\b");

        if (tf >= 0)
        {
          cs1 = cs.substr(tf+1);
          cs =  cs.substr(0, tf);
          if (!bEnab)
          {
            if (!bSelect)
            {
              TRect rcText1(rcText);
              rcText1.Inflate(-1,-1);
              dc.SetTextColor(TColor::Sys3dHilight);
              dc.DrawText(cs.c_str(),  -1, rcText1, DT_VCENTER|DT_LEFT);
              dc.DrawText(cs1.c_str(), -1, rcText1, DT_VCENTER|DT_RIGHT);
              dc.SetTextColor(TColor::SysGrayText);
              dc.DrawText(cs.c_str(),  -1, rcText, DT_VCENTER|DT_LEFT);
              dc.DrawText(cs1.c_str(), -1, rcText, DT_VCENTER|DT_RIGHT);
            }
            else
            {
              dc.SetTextColor(TColor::SysMenu);
              dc.DrawText(cs.c_str(),  -1, rcText, DT_VCENTER|DT_LEFT);
              dc.DrawText(cs1.c_str(), -1, rcText, DT_VCENTER|DT_RIGHT);
            }
          }
          else
          {
            dc.DrawText(cs.c_str(),  -1, rcText, DT_VCENTER|DT_LEFT);
            dc.DrawText(cs1.c_str(), -1, rcText, DT_VCENTER|DT_RIGHT);
          }
        }
        else
        {
          if (!bEnab)
          {
            if (!bSelect)
            {
              TRect rcText1(rcText);
              rcText1.Inflate(-1,-1);
              dc.SetTextColor(TColor::Sys3dHilight);
              dc.DrawText(cs.c_str(), -1,	rcText1, DT_VCENTER|DT_LEFT|DT_EXPANDTABS);
              dc.SetTextColor(TColor::SysGrayText);
              dc.DrawText(cs.c_str(), -1, rcText, DT_VCENTER|DT_LEFT|DT_EXPANDTABS);
            }
            else
            {
              dc.SetTextColor(TColor::SysMenu);
              dc.DrawText(cs.c_str(), -1,	rcText, DT_VCENTER|DT_LEFT|DT_EXPANDTABS);
            }
          }
          else
            dc.DrawText(cs.c_str(), -1, rcText, DT_VCENTER|DT_LEFT|DT_EXPANDTABS);
        }

        dc.SetTextColor(ocr);
        dc.SetBkMode(obk);
      }
			dc.RestoreFont();
      delete pft;
		}
		res = true;
	}
	return res;
}

bool TBmpMenu::MeasureItem(MEASUREITEMSTRUCT far& MIS)
// To be called from the TWindow EvMeasureItem(..)
// Call the base class if it return false
{
	bool res = false;
	if (MIS.CtlType == ODT_MENU)
	{
		TBmpMenuItem* pItem = (TBmpMenuItem*)MIS.itemData;
		if (pItem)
		{
			if (pItem->Cmd == -3) // is a separator
			{
				MIS.itemWidth  = 10;
				MIS.itemHeight = 6;
			}
			else
      {
        string cs(pItem->Text);
        if (cs != "")
        {
          TClientDC dc(0);

          // get the actual windows wide menu font
          //
          TFont* pft;
          NONCLIENTMETRICS nm;
          nm.cbSize = sizeof (NONCLIENTMETRICS);
          if (SystemParametersInfo(SPI_GETNONCLIENTMETRICS, nm.cbSize, &nm, 0))
            pft = new TFont(&nm.lfMenuFont);
          else
            pft = new TFont((HFONT)GetStockObject(DEFAULT_GUI_FONT));
          CHECK(pft);
          dc.SelectObject(*pft);

          if (pItem->Cmd == -4) // is a title item
          {
            TRect rci(0,0,0,0);
            dc.DrawText(cs.c_str(), -1, rci, DT_CALCRECT|DT_TOP|DT_VCENTER|DT_SINGLELINE);
            MIS.itemHeight = rci.Height();
            MIS.itemWidth = rci.Width();
          }
          else
          {
            int h = GetSystemMetrics(SM_CYMENU);
            MIS.itemHeight = h>ImageSize.cy+3 ? h : ImageSize.cy+3;

            TSize sz = dc.GetTabbedTextExtent(cs.c_str(), strlen(cs.c_str()), 0, 0);
            MIS.itemWidth = MIS.itemHeight + 2 + sz.cx;

            // make the menu bigger, because sometimes the calculated
            // width is not enough
            //
            MIS.itemWidth += 2 * (sz.cx/strlen(cs.c_str()));
          }

          dc.RestoreFont();
          delete pft;
        }
        else
        {
          MIS.itemHeight = ImageSize.cy + 3;
          MIS.itemWidth  = 100;
        }
      }
		}
		res = true;
	}
	return res;
}

bool TBmpMenu::FindKeyboardShortcut(UINT nChar, UINT, HMENU hMenu, LRESULT& lRes)
// To be called from the TWindow EvMenuChar(..)
// Call the base class if it return false
{
	PRECONDITION(hMenu);
	int nItem = ::GetMenuItemCount(hMenu);
	string csChar((TCHAR)nChar);
	csChar.to_lower();
	while ((--nItem)>=0)
	{
		UINT itemId = ::GetMenuItemID(hMenu, nItem);
		if (itemId != 0)
		{
      TMenuItemInfo mii(MIIM_DATA|MIIM_TYPE);
			if (::GetMenuItemInfo(hMenu, nItem, TRUE, &mii))
			{
				if (mii.fType&MFT_OWNERDRAW)
				{
					TBmpMenuItem* pItem = (TBmpMenuItem*)mii.dwItemData;
					if (pItem)
					{
						string csItem(pItem->Text);
						csItem.to_lower();
						int iAmperIdx = csItem.find("&");
						if (iAmperIdx >= 0)
						{
							csItem = csItem.substr(iAmperIdx + 1, 1);
							if (csItem == csChar)
							{
								lRes = MAKELONG((WORD)nItem, 2);
								return true;
							}
						}
					}
				}
			}
		}
	}
	return false;
}

int TBmpMenu::PreGetHintText(HMENU hmenu, uint id, LPTSTR buf, int size,
                          THintText hintType)
// Retrieves the hint text associated with a particular Id
// To be called from DecoratedFrame GetHintText(...)
// Call the base class if it return 0.
{
  int numBytes = 0;

  if (hintType == htTooltip)
  {
    if (hmenu && ::IsMenu(hmenu))
    {
	    // For tooltips, we first look if the same id exists on the
  	  // bitmap menu and use the associated menu string.
    	//
      bool found = false;
      bool accValid = true;
      for (UINT i=0; i<MenuItems.GetItemsInContainer(); i++)
      {
    		if (MenuItems[i]->Cmd == (int)id)
        {
          if (strlen(MenuItems[i]->Text) < (uint)size)
	          strcpy(buf, MenuItems[i]->Text);
          else
          {
            strncpy(buf, MenuItems[i]->Text, size-1);
            buf[size-1] = 0;
            accValid = false;
          }
          found = true;
          break;
			  }
      }
      if (found == false)
        return numBytes;

      // Trim accelerator text e.g., '\tCtrl+X' (if any)
      // Accelerator can be \b instead of \t (NOTE: \a in .rc make a \b in res)
      //
      LPTSTR acc = new TCHAR[size];
      *acc = 0;
      LPTSTR c = strchr(buf, '\t');
      if (c == 0)
        c = strchr(buf, '\b');
      if (c)
      {
        if (bShowAccInTooltips && accValid)
        {
          // save the accelerator string for showing in tooltips
          //
          strcpy(acc, c+1);
        }
        *c = 0;
      }

      // Trim trailing dots, e.g., '...' (if any)
      //
      int i;
      for (i = strlen(buf) - 1; i >= 0; i--)
        if (buf[i] == '.')
          buf[i] = 0;
        else
          break;

      // Eliminate '&' (just emliminate first one)
      //
      i = -1;
      while(buf[++i])
        if (buf[i] == '&') {
          do {
            buf[i] = buf[i+1];
            i++;
          } while (buf[i]);
          break;
        }

      // Show the accelerator in tooltip
      //
      if (bShowAccInTooltips && accValid && *acc != 0 &&
          (strlen(buf) + strlen(acc) + 3) < (uint)size)
      {
   	    strcat(buf, " (");
     	  strcat(buf, acc);
       	strcat(buf, ")");
      }

      delete[] acc;

      numBytes = strlen(buf);
    }
  }
  return numBytes;
}

bool TBmpMenu::SetBackBitmap(unsigned int resId)
// Set the background bitmap. return true if successful
{
	PRECONDITION(Module);
  delete BackBitmap;
  BackBitmap = 0;
  bBackBitmap = false;
  if (resId > 0) {
	  BackBitmap = new TBitmap(Module->LoadBitmap(TResId(resId)));
 		if (BackBitmap && (HBITMAP)*BackBitmap)
	  	bBackBitmap = true;
  }
  return bBackBitmap;
}

void TBmpMenu::ExcludeBitmap(unsigned resId)
// Exclude the bitmap with resId. It will not be shown in menus
{
#if (OWLInternalVersion >= 0x06000000L)
  if (ExcludedBitmaps.Find(resId) == NPOS)
    ExcludedBitmaps.Add(resId);
#else
  if (ExcludedBitmaps.Find(&resId) == INT_MAX)
    ExcludedBitmaps.Add(new unsigned int(resId));
#endif
}

void TBmpMenu::IncludeBitmap(unsigned resId)
// Include the bitmap with resId. By default all bitmaps in module
// will be shown in menus
{
#if (OWLInternalVersion >= 0x06000000L)
  uint idx = ExcludedBitmaps.Find(resId);
  if (idx != NPOS)
#else
  uint idx = ExcludedBitmaps.Find(&resId);
  if (idx != INT_MAX)
#endif
    ExcludedBitmaps.Destroy(idx);
}

bool TBmpMenu::SetMenuItemText(HMENU hMenu, UINT pos, LPCTSTR txt)
// Set the owner draw menu item text like Win API function:
//   ModifyMenu(hMenu, pos, MF_BYPOSITION | MF_STRING, .., txt)
{
  PRECONDITION(hMenu);
  TCHAR str[MaxMenuStringLen];
  strcpy(str, (txt) ? txt : "");

  TMenuItemInfo mii(MIIM_TYPE | MIIM_DATA | MIIM_STATE | MIIM_ID | MIIM_SUBMENU);
	if (::GetMenuItemInfo(hMenu, pos, true, &mii))
  {
		if (mii.fType&MFT_OWNERDRAW && mii.dwItemData != 0)
    {
	    // search menu item and change the text
  	  //
    	for (UINT i=0; i<MenuItems.GetItemsInContainer(); i++)
	    {
  	    if (MenuItems[i]->Menu == hMenu && MenuItems[i]->Pos == pos)
	      {
          if (MenuItems[i]->Cmd > 0)
          {
            if ((UINT)MenuItems[i]->Cmd == mii.wID)
            {
		          if (strcmp (MenuItems[i]->Text, str) == 0)
					      return true; // nothing to change
	  		      strcpy(MenuItems[i]->Text, str);
  	  		    return ::ModifyMenu(hMenu, pos, MF_BYPOSITION|MF_OWNERDRAW|mii.fState,
		    	      (LPARAM)mii.wID, (LPCTSTR)MenuItems[i]);
            }
          }
          else
          {
            if (mii.hSubMenu != 0)
            {
		          if (strcmp (MenuItems[i]->Text, str) == 0)
					      return true; // nothing to change
	  		      strcpy(MenuItems[i]->Text, str);
  	  		    return ::ModifyMenu(hMenu, pos, MF_BYPOSITION|MF_OWNERDRAW|mii.fState,
		    	      (LPARAM)-1, (LPCTSTR)MenuItems[i]);
            }
          }
        }
      }
    }
  }
  return false;
}

void TBmpMenu::RemapMenuRecursiv(HMENU hMenu)
{
	static int iRecurse = 0;
	iRecurse ++;

	int nItem = ::GetMenuItemCount(hMenu);
	while ((--nItem)>=0)
	{
		UINT itemId = ::GetMenuItemID(hMenu, nItem);
		if (itemId == (UINT) -1)
		{
			HMENU pops = ::GetSubMenu(hMenu, nItem);
			if (pops)
      	RemapMenuRecursiv(pops);
			if (bIsPopup || iRecurse>0)
			{
				TCHAR cs[MaxMenuStringLen];
				if (::GetMenuString(hMenu, nItem, cs, MaxMenuStringLen, MF_BYPOSITION) > 0)
				{
					TBmpMenuItem* pItem = AddMenuItem(hMenu, cs,
          	(bIsPopup == false && iRecurse == 1)?-4:-2, nItem);
					::ModifyMenu(hMenu, nItem, MF_BYPOSITION|MF_OWNERDRAW,
          	(UINT)-1, (LPCTSTR)pItem);
				}
			}
		}
		else
		{
      TBmpMenuItem* pItem;
			TCHAR cs[MaxMenuStringLen];
		  UINT oldState = ::GetMenuState(hMenu, nItem, MF_BYPOSITION);
			if (!(oldState&MF_OWNERDRAW) && !(oldState&MF_BITMAP))
      {
        if ((oldState&MF_SEPARATOR) || itemId == 0)
        {
          // Insert separator items only if a background bitmap is drawn
          //
          if (bBackBitmap)
          {
            pItem = AddMenuItem(hMenu, "--", -3, nItem);
						::ModifyMenu(hMenu, nItem, MF_BYPOSITION|MF_OWNERDRAW|oldState,
	           	(LPARAM)itemId, (LPCTSTR)pItem);
          }
        }
        else
				{
					if (::GetMenuString(hMenu, nItem, cs, MaxMenuStringLen, MF_BYPOSITION) > 0)
    	    {
      	    TryToLoadBmpResource(itemId);
        	  pItem = AddMenuItem(hMenu, cs, itemId, nItem);
		  			::ModifyMenu(hMenu, nItem, MF_BYPOSITION|MF_OWNERDRAW|oldState,
	           	(LPARAM)itemId, (LPCTSTR)pItem);
  	      }
        }
      }
    }
	}
	iRecurse --;
}

void TBmpMenu::DrawDisabledButton(TDC& dc, const TRect& rc)
{
	// create a monochrome memory DC
  //
	TMemoryDC ddc;
	TBitmap bmp(ddc, rc.Width(), rc.Height());
	ddc.SelectObject(bmp);

	// build a mask
  //
	ddc.PatBlt(0, 0, rc.Width(), rc.Height(), WHITENESS);
	dc.SetBkColor(TColor::Sys3dFace);
	ddc.BitBlt(0, 0, rc.Width(), rc.Height(), dc, rc.left, rc.top, SRCCOPY);
	dc.SetBkColor(TColor::Sys3dHilight);
	ddc.BitBlt(0, 0, rc.Width(), rc.Height(), dc, rc.left, rc.top, SRCPAINT);

	// Copy the image from the toolbar into the memory DC
	// and draw it (grayed) back into the toolbar.
  //
	dc.FillRect(rc, TBrush(TColor::Sys3dFace));
	dc.SetBkColor(RGB(0, 0, 0));
	dc.SetTextColor(RGB(255, 255, 255));
	TBrush brShadow(TColor::Sys3dShadow);
  TBrush brHilight(TColor::Sys3dHilight);
	dc.SelectObject(brHilight);
	dc.BitBlt(rc.left+1, rc.top+1, rc.Width(), rc.Height(), ddc, 0, 0, 0x00E20746L);
	dc.SelectObject(brShadow);
	dc.BitBlt(rc.left, rc.top, rc.Width(), rc.Height(), ddc, 0, 0, 0x00E20746L);

	// reset DCs
  //
	dc.RestoreBrush();
  dc.RestoreBrush();
	ddc.RestoreBitmap();
}

void TBmpMenu::DrawCheckmark(TDC& dc, int x, int y, COLORREF color)
{
// ......X
// .....XX
// X...XXX
// XX.XXX.
// XXXXX..
// .XXX...
// ..X....

  dc.SetPixel(x,y+2,color);
  dc.SetPixel(x,y+3,color);
  dc.SetPixel(x,y+4,color);

  dc.SetPixel(x+1,y+3,color);
  dc.SetPixel(x+1,y+4,color);
  dc.SetPixel(x+1,y+5,color);

  dc.SetPixel(x+2,y+4,color);
  dc.SetPixel(x+2,y+5,color);
  dc.SetPixel(x+2,y+6,color);

  dc.SetPixel(x+3,y+3,color);
  dc.SetPixel(x+3,y+4,color);
  dc.SetPixel(x+3,y+5,color);

  dc.SetPixel(x+4,y+2,color);
  dc.SetPixel(x+4,y+3,color);
  dc.SetPixel(x+4,y+4,color);

  dc.SetPixel(x+5,y+1,color);
  dc.SetPixel(x+5,y+2,color);
  dc.SetPixel(x+5,y+3,color);

  dc.SetPixel(x+6,y,color);
  dc.SetPixel(x+6,y+1,color);
  dc.SetPixel(x+6,y+2,color);
}

void TBmpMenu::DrawRadioDot(TDC& dc, int x, int y, COLORREF color)
{
  TRect rcDot(x, y, x+7, y+7);
  TBrush brush(color);
  TPen pen(color,0);
  dc.SelectObject(brush);
  dc.SelectObject(pen);
  dc.Ellipse(rcDot);
  dc.RestoreBrush();
  dc.RestorePen();
}

int TBmpMenu::FindImageItem(const int cmd)
{
	for (UINT i=0; i<ImageItems.GetItemsInContainer(); i++)
		if (ImageItems[i]->Cmd == cmd)
    	return ImageItems[i]->ImageIdx;

	return -1;
}

TBmpMenuItem* TBmpMenu::AddMenuItem(HMENU hMenu, LPCTSTR txt, int cmd, UINT pos)
{
  PRECONDITION(hMenu);

  TCHAR str[MaxMenuStringLen];
  if (txt)
    strcpy(str, txt);
  else
	  strcpy(str, "");

  // if the same item exist, do not insert the menu Item
  //
  for (UINT i=0; i<MenuItems.GetItemsInContainer(); i++)
  {
		if (MenuItems[i]->Menu == hMenu &&
    		MenuItems[i]->Cmd == cmd &&
        MenuItems[i]->Pos == pos &&
        strcmp(MenuItems[i]->Text, txt) == 0)
      return MenuItems[i];
  }

	TBmpMenuItem* pItem = new TBmpMenuItem;
  CHECK(pItem);

  // set the menu item data
  //
  pItem->Menu = hMenu;
  pItem->Pos = pos;
	pItem->Cmd = cmd;
	if (cmd >= 0)
  	pItem->ImageIdx = FindImageItem(cmd);
	else
  	pItem->ImageIdx = cmd;

	strcpy(pItem->Text, str);

  MenuItems.Add(pItem);
	return pItem;
}

void TBmpMenu::AddImageItem(int idx, WORD cmd)
{
  if (idx < 0)
  	return;

	TBmpMenuImageItem* pItem = new TBmpMenuImageItem;
  CHECK(pItem);

	pItem->Cmd = (int) cmd;
	pItem->ImageIdx = idx;

  ImageItems.Add(pItem);
}

bool TBmpMenu::TryToLoadBmpResource(unsigned int resId)
{
	PRECONDITION(Module);

  // Check if the bitmap is excluded
  //
#if (OWLInternalVersion >= 0x06000000L)
  if (ExcludedBitmaps.Find(resId) != NPOS)
#else
  if (ExcludedBitmaps.Find(&resId) != INT_MAX)
#endif
    return false;

  // add only one bmp with resId to the image list
  //
  if (FindImageItem(resId) < 0)
  {
    // load bitmap from module and add them to image list
    //
  	TBitmap srcBmp(Module->LoadBitmap(TResId(resId)));
  	if ((HBITMAP)srcBmp != 0)
    {
		 	int nBmpItems = ImageList->Add(srcBmp, RGB(192,192,192));
			AddImageItem(nBmpItems, (WORD)resId);
      return true;
    }
  }
	return false;
}

void TBmpMenu::SyncronizeAllMenuItems()
// Syncronize the intern menu items and the real windows menus and items
{
	UINT i, num=MenuItems.GetItemsInContainer();
  TBmpMenuItem* item;
  for (i=num; i>0; i--)
  {
  	// Delete a intern menu item, if the hMenu member
		// is not valid (e.g. the associated menu is destroyed).
    //
    item = MenuItems[i-1];
    if (::IsMenu(item->Menu) == false)
      MenuItems.Destroy(i-1);
  }

  num=MenuItems.GetItemsInContainer();
  for (i=num; i>0; i--)
  {
    bool bRet, bDestroy = false;
    TMenuItemInfo mii(MIIM_TYPE | MIIM_DATA | MIIM_STATE | MIIM_ID | MIIM_SUBMENU);
    item = MenuItems[i-1];
    if (item->Cmd > 0)
      bRet = ::GetMenuItemInfo(item->Menu, item->Cmd, false, &mii);
    else
      bRet = ::GetMenuItemInfo(item->Menu, item->Pos, true, &mii);
    if (bRet == false)
    {
	    // if windows menu item not present (e.g. removed), delete intern item
			//
		  MenuItems.Destroy(i-1);
    }
    else
    {
      if (item->Cmd > 0) // intern item is a simple item
      {
        if (mii.hSubMenu != 0)
        {
          // if windows menu item is a submenu, delete intern item
          //
			    bDestroy = true;
        }
        else if ((UINT)item->Cmd != mii.wID)
        {
          // if windows command change, delete intern item
          //
			    bDestroy = true;
        }
      }
      else
      {
        if (item->Cmd == -3)
        {
          // if intern item is a separator and there is no background bitmap,
					// delete intern item
          //
          if (bBackBitmap == false)
				    bDestroy = true;
        }
        else if (mii.hSubMenu == 0)
        {
          // if intern item is a submenu and windows menu is a simple item,
          // delete intern item
          //
			    bDestroy = true;
        }
      }

 			if (bDestroy == false &&
      		!(mii.fType&MFT_OWNERDRAW)) // windows item is not ownerdrawn
      {
        if ((mii.fType&MFT_STRING) && mii.dwTypeData != 0)
        {
          // if windows item is a string item, try to syncronize the items
          //
		      strcpy(item->Text, mii.dwTypeData);
  	      UINT cmd = (item->Cmd <= 0) ? (UINT)-1 : item->Cmd;
  		    ::ModifyMenu(item->Menu, item->Pos,
						MF_BYPOSITION|MF_OWNERDRAW|mii.fState, cmd, (LPCTSTR)item);
        }
      }

      // if windows item is ownerdrawn and have no valid datas, delete itern item
      //
      if ((mii.fType&MFT_OWNERDRAW) && mii.dwItemData == 0)
		    bDestroy = true;

      // check menu type before destroy the intern item
      //
      if (bDestroy && !(mii.fType&MFT_OWNERDRAW))
			  MenuItems.Destroy(i-1);
    }
  }
}

#endif
